#!/usr/bin/python
# -*- coding: utf-8 -*-

from snapconnect import snap
from time import time
from time import sleep
import logging
import numpy
import ttyusb
import sys
import random
import string
import binascii

timeout_local_address = 1
timeout_set_pin_dir = 1
timeout_write_pin = 2


def randomword(length):
    return ''.join(random.choice(string.lowercase) for i in range(length))


class SnapEngine(snap.Snap):

    def __init__(
        self,
        port=None,
        expected_address=None,
        timeout=3,
        ports=None,
        level=logging.ERROR,
        addr='\x00\x00\x20',
        license_file='License.dat',
            nvparams_file='nvparams.dat'):
        super(SnapEngine, self).__init__(
            funcs={},
            license_file=license_file,
            nvparams_file=nvparams_file,
            addr=addr
            )
        # self.function_destination = None
        logging.basicConfig(level=level)  # messages to STDOUT
        self.retry = 15
        self.timeout = timeout
        # self.add_rpc_func('callback_function', self.callback_function)
        if port:
            print 'opening:', port
            self.open_serial(snap.SERIAL_TYPE_RS232, port)
        elif ports:
            for port in ports:
                print 'opening:', port
                self.open_serial(snap.SERIAL_TYPE_RS232, port)
        if expected_address:
            self.local_address = self.call(None, 'localAddr')
            if expected_address != self.local_address:
                raise Exception(
                    'ERROR: expected address is %s, but found %s on port %s',
                    (expected_address, self.local_address, port))
        # sys.stderr.write('initialized_eng_local_addr: %s\n' %
            # binascii.hexlify(self.local_addr()))

    def call(self, dest, function_name, *args):
        cb_name = randomword(10)
        self.add_rpc_func(cb_name, self.callback_function)
        try:
            return self.call1(cb_name, dest, function_name, *args)
        finally:
            del self.local_dispatcher.funcs[cb_name]

    def call1(self, cb_name, dest, function_name, *args):
        # sys.stderr.write('call %s%s\n' % (function_name, repr(args)))
        # sys.stderr.write(str(self) + '\n')
        self.response_received = None
        self.function_destination = dest
        self.number_of_requests = 0
        while True:
            if self.number_of_requests > self.retry:
                raise Exception('ERROR: timeout calling %s' % function_name)
            self.number_of_requests += 1
            if self.number_of_requests > 1:
                # print 'request #%d calling %s' % (
                    # self.number_of_requests, function_name)
                if self.function_destination:
                    sys.stderr.write(
                        'request #%d: %s is calling: rpc(%s, %s, ...)\n' %
                        (
                            self.number_of_requests,
                            binascii.hexlify(self.local_addr()),
                            binascii.hexlify(self.function_destination),
                            function_name
                        )
                    )
            if self.function_destination:
                # sys.stderr.write(
                    # '%s calling: rpc %s\n' %
                    # (binascii.hexlify(self.local_addr()),
                        # binascii.hexlify(self.function_destination)))
                self.rpc(
                    self.function_destination,
                    'callback',
                    cb_name,
                    function_name, *args)
            else:
                # sys.stderr.write('mcast_rpc...\n')
                self.mcast_rpc(
                    1, 1, 'callback', cb_name,
                    function_name, *args)
            self.request_sent = time()
            function_response_timeout = self.timeout + self.request_sent
            while True:
                if time() >= function_response_timeout:
                    break
                self.poll()
                if self.response_received:
                    self.response_time = \
                        self.response_received - self.request_sent
                    # print 'response_time =', self.response_time
                    return self.return_value

    def print_info(self, myid, macinfo):
        print 'engine: ', myid
        print 'name: ', self.call(macinfo, 'loadNvParam', 8)
        local_address = self.call(macinfo, 'localAddr')
        local_address_str = (
            binascii.hexlify(local_address) if local_address else ''
            )
        print 'localAddr: ', local_address_str
        print 'channel: ', self.call(macinfo, 'loadNvParam', 4)
        nid = self.call(macinfo, 'loadNvParam', 3)
        nid_str = ('0x%.2x%.2x' % ((nid >> 8) % 256, nid % 256)) if nid else ''
        print 'networkId: ', nid_str
        fea = self.call(macinfo, 'loadNvParam', 11)
        fea_str = ('0x%.2x%.2x' % ((fea >> 8) % 256, fea % 256)) if fea else ''
        print 'feature: ', fea_str
        print

    def callback_function(self, return_value):
        # sys.stderr.write('callback_function return_value: %s\n' %
        # repr(return_value))
        # sys.stderr.write(str(self) + '\n')
        # sys.stderr.write('local_addr: %s\n' %
            # binascii.hexlify(self.local_addr()))
        if (
            self.function_destination is None or
                self.rpc_source_addr() == self.function_destination):
            self.response_received = time()
            self.return_value = return_value

    def read_voltage(self, mac, channel, poly, repeat=25):
        adcs = []
        while len(adcs) < repeat:
            adc = self.call(mac, 'readAdc', channel)
            # print adc
            adcs.append(float(adc))
        adc = sum(adcs) / len(adcs)
        # print 'adc =', adc
        return numpy.polyval(poly, adc)

    def open_serial(
        self,
        serial_type=snap.SERIAL_TYPE_RS232,
        port=None,
        vid=None,
        pid=None,
        serial_number=None,
        reconnect=False,
        dll_path=None,
        forward_groups=None,
        baudrate=None,
        io_loop=None,
            close_all=True):
        if port is None:
            if vid is None:
                raise ValueError("vid cannot be None if port is None")
            if pid is None:
                raise ValueError("pid cannot be None if port is None")
            port = ttyusb.find_tty_usb_sn(vid, pid, serial_number)
        if close_all:
            self.close_all_serial()
        if dll_path is not None:
            super(SnapEngine, self).open_serial(
                serial_type,
                port,
                reconnect=reconnect,
                dll_path=dll_path,
                forward_groups=forward_groups,
                baudrate=baudrate,
                io_loop=io_loop
                )
        else:
            super(SnapEngine, self).open_serial(
                serial_type,
                port,
                reconnect=reconnect,
                forward_groups=forward_groups,
                baudrate=baudrate,
                io_loop=io_loop
                )

    def open_usb_serial(
        self,
        usb_serial,
        reconnect=False,
        dll_path=None,
        forward_groups=None,
        baudrate=None,
        io_loop=None,
            close_all=True):
        self.open_serial(
            vid=usb_serial['vid'],
            pid=usb_serial['pid'],
            serial_number=usb_serial['sn'],
            reconnect=reconnect,
            dll_path=dll_path,
            forward_groups=forward_groups,
            baudrate=baudrate,
            io_loop=io_loop,
            close_all=close_all,
        )

    def read_snappy_os_version(self, mac):
        version_major = self.call(mac, 'getInfo', 5)
        version_minor = self.call(mac, 'getInfo', 6)
        version_build = self.call(mac, 'getInfo', 7)
        return (version_major, version_minor, version_build)

    def enable_crc(self, mac, try_enabled_first=False, enable=True):
        mac_str = 'Unknown' if mac is None else binascii.hexlify(mac)
        retry_save = self.retry
        self.retry = 1
        comm_retry = 0
        while True:
            comm_retry += 1
            if comm_retry > retry_save:
                self.retry = retry_save
                raise Exception(
                    'communication failed with node: %s' % mac_str
                    )
            if try_enabled_first:
                try:
                    self.save_nv_param(11, 0x100)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0xfeff & nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0)
                    break
                except:
                    print(
                        'with CRC enabled, communication failed with node %s' %
                        mac_str
                        )
                try:
                    self.save_nv_param(11, 0)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if not enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0x100 | nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0x100)
                    break
                except:
                    print(
                        'with CRC disabled, communication failed with node %s' %
                        mac_str
                        )
            else:
                try:
                    self.save_nv_param(11, 0)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if not enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0x100 | nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0x100)
                    break
                except:
                    print(
                        'with CRC disabled, communication failed with node %s' %
                        mac_str
                        )
                try:
                    self.save_nv_param(11, 0x100)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0xfeff & nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0)
                    break
                except:
                    print(
                        'with CRC enabled, communication failed with node %s' %
                        mac_str
                        )
        self.retry = retry_save

    def rpcpoll(self, dest, function_name, *args):
        self.rpc(dest, function_name, *args)
        self.poll()

if __name__ == "__main__":
    ports = ttyusb.find_tty_usb('067b', '2303')
    print ports
    port = ports[0]
    eng = SnapEngine(port)
    eng.enable_crc('\x05\x46\x52')
    print eng.read_snappy_os_version('\x05\x46\x52')
